/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.util.*;

public class Cache extends Thread {
	
	Node node;
	
	Sensor sensor;
	
	String attribute;
	
	int SP;	
	double E;	
	int SPAN;	
	long IT;
	LinearRegressionStructure lrs = new LinearRegressionStructure(); // R
	
	Vector methods = new Vector(10,5);
	
	Vector data = new Vector (100,50);
	
	long currentTimestamp;  // unit: second
	double currentReading;
	
	int size = 0;  // current size of the cache, unit: byte
	
	boolean stopped = false;  // indicate whether this cache has been stopped
	
	double cost1, cost2;
	CachePoint s1, s2, s3, s4;
	int markf, markl;
	double cf [], cl [];
	double maxf, avgf, minf;
	double maxl, avgl, minl;
	double minCost;		
	
	Cache (Node node, String attribute)
	  {
	  	this.node = node;
	  	
	  	this.attribute = attribute;
	  	
	  	int i;
	  	for (i=0;i<node.sensors.size();i++)
	  	  {
	  	  	Sensor sensor = (Sensor)node.sensors.elementAt(i);
	  	  	if (sensor.attribute.equals(attribute))
	  	  	  {this.sensor = sensor; break;}
	  	  }
	    if (this.sensor == null)
	      {
	      	System.err.println("Unknown sensor type for a cache. Please check.");
	      	System.exit(0);
	      }
	  	
	  	currentTimestamp = 0;	  	
	  }
	  
	public void run()
	  {
	  	while (true)
	  	  {
	  	    if (node.caches.size() < Educt.numQueries)
	  	      continue;
	  	    
	  	    String line = sensor.sample(currentTimestamp, SP);
	  	    
	  	    if (line == null)
	  	      {
	  	      	stopped = true;;
	  	      	
	  	      	int i;
	  	      	for (i=0;i<node.caches.size();i++)
	  	      	  if (!((Cache)node.caches.elementAt(i)).stopped)
	  	      	      break;
	  	      	if (i == node.caches.size())
	  	      	  node.educt.nodeTermination(node);
	  	      	 
	  	      	return;
	  	      }
	  	      
	  	    int mark = line.indexOf(" ");
	  	    if (mark == -1)
	  	      mark = line.indexOf("\t");
	  	    String str = line.substring(0,mark);
	  	    currentTimestamp = Long.parseLong(str);  // the timestamp format in the data file is an integer starting from 0 with unit second
	  	    str = line.substring(mark);
	  	    currentReading = Double.parseDouble(str);
	  	    
	  	    if (methods.size() != 0)
	  	      node.cm.newReadingArrival(this,currentReading);
	  	    else
	  	      for (int i=0;i<node.queries.size();i++)
	  	        {
	  	          Query q = (Query)node.queries.elementAt(i); 
	  	          
	  	          if ((q.type == Query.THRESHOLD) && ((Predicate)q.predicates.elementAt(0)).attribute.equals(this.attribute))
	  	            {
	  	          	  long startTime = System.currentTimeMillis();
	  	          	  RP p = new RP();
	  	          	  p.value = currentReading;
	  	          	  p.timestamp = currentTimestamp;
	  	          	  node.executeQuery(p, (Query)node.queries.elementAt(i));
	  	          	  long endTime = System.currentTimeMillis();
	  	          	  q.singleCT = endTime-startTime;
	  	          	  if (q.singleCT > q.maxCT)
	  	          	    q.maxCT=q.singleCT;	  	          	  
	  	          	  q.sumCT += q.singleCT;
	  	          	  if (q.singleCT < q.minCT)
	  	          	    q.minCT = q.singleCT;
	  	          	  	
	  	          	  q.maxSR =  Educt.ATTR_SIZE;
	  	          	  q.sumSR += Educt.ATTR_SIZE;
	  	          	  q.minSR = Educt.ATTR_SIZE;
	  	          	  
	  	          	  q.count++;        	
	  	            }
	  	        }  	    
	  	  }
	  	  	  	
	  }  //run	
}