/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.util.*;

public class CacheManager {
	
	Node node;	
	
	static final boolean PCA = true;
	static final boolean PLR = true;
	
	boolean isPPFP;  // whether the new reading is a PPFR for some jump or spike method		
	PPFP ppfp;
	
	CacheManager (Node node)
	  {
	  	this.node = node;	  	
	  }
	  
	public void newReadingArrival(Cache C, double reading)
	  {
	  	int i,j;
	  	long startTime, endTime;
	  	
	  	startTime = System.currentTimeMillis();
	  	RP s = new RP();
	  	s.timestamp = C.currentTimestamp;
	  	s.value = reading;
	  	endTime = System.currentTimeMillis();	  	
	  	
	  	Query q = null;
	  	for (i=0;i<node.queries.size();i++)
	  	  if ((q=(Query)node.queries.elementAt(i)).M.attribute.equals(C.attribute))
	  	  	q.singleCT = endTime - startTime;
	  		  	
	    isPPFP = false;
	    ppfp = null;
	    
	  	for (i=0;i<C.methods.size();i++)
	  	  {
	  	  	EDMethod M = (EDMethod)C.methods.elementAt(i);	  	  	
	  	  	
	  	  	for (j=0;j<node.queries.size();j++)
	  	  	  if (M.q.equals(q=(Query)node.queries.elementAt(j)))
	  	  	    break;
	  	  	startTime = System.currentTimeMillis();
	  	  	
	  	  	if (M.type == EDMethod.PATTERN)
	  	  	  generalPatternMatching((PatternMethod)M, C, s);
	  	  	else
	  	  	  if (M.type == EDMethod.HORIZON)
	  	  	    horizonPatternMatching((HorizonMethod)M, C, s); 
	  	  	  else
	  	  	    if (M.type == EDMethod.SLOPE)
	  	  	      slopePatternMatching((SlopeMethod)M, C, s);
	  	  	    else
	  	  	      if (M.type == EDMethod.OSCILLATION)
	  	  	        oscillationPatternMatching((OscillationMethod)M, C, s);
	  	  	      else
	  	  	        if (M.type == EDMethod.JUMP)
	  	  	          jumpPatternMatching((JumpMethod)M, C, s);
	  	  	        else
	  	  	          if (M.type == EDMethod.SPIKE)
	  	  	            spikePatternMatching((SpikeMethod)M, C, s);	  	  	          
	  	  	
	  	  	endTime = System.currentTimeMillis();
	  	  	q.singleCT += endTime - startTime; 	  	            
	  	  }
	  	  
	  	startTime = System.currentTimeMillis();
	  	updateCache(C, s);
	  	endTime = System.currentTimeMillis();  		  	
	  	
	  	for (i=0;i<node.queries.size();i++)
	  	  if ((q=(Query)node.queries.elementAt(i)).M.attribute.equals(C.attribute))
	  	    {
	  	  	  q.singleCT += endTime - startTime;	
	  	  	  
	  	  	  if (q.singleCT > q.maxCT)
	  	        q.maxCT = q.singleCT;	  	      
	  	   	  q.sumCT += q.singleCT;
	  	      if (q.singleCT < q.minCT)
	  	        q.minCT = q.singleCT;
	  	        
	  	      q.count++;
	  	      
	  	      if (C.size > q.maxSR)
	  	        q.maxSR = C.size;
	  	      q.sumSR += C.size;
	  	      if (C.size < q.minSR)
	  	        q.minSR = C.size;     
	  	    }
	  }
	 
	public int findMark(Cache C, CachePoint s)
	  {
	  	int i;
	  	CachePoint ss;
	  	for (i=0;i<C.data.size();i++)
	  	  {
	  	  	ss = (CachePoint)C.data.elementAt(i);
	  	  	if (ss==s)
	  	  	  break;
	  	  }
	  	  
	  	if (i==C.data.size()) return -1;
	  	else return i;
	  	
	  }  //findMark
	  
	public long getStartTimestamp(Cache C, CachePoint s)
	  {
	  	int i = findMark(C,s);
	  	
	  	if (i==-1)
	  	  {
	  	  	return -1;
	  	  	
	  	  	//System.err.println("Function getStartTimestamp: cannot find the cache point");
	  	  	//System.err.println("timestamp = " + s.timestamp);
	  	  	//System.exit(1);
	  	  }
	  	  
	  	if (i==0) return C.IT;
	  	else return ((CachePoint)C.data.elementAt(i-1)).timestamp + C.SP;
	  	
	  }  //getStartTimestamp
	  
	public int findCacheItem(Cache C,long t)
	  {
	  	CachePoint s;
	  	int i;
	  	long ts, te;
	  	for (i=0;i<C.data.size();i++)
	  	  {
	  	  	s = (CachePoint)C.data.elementAt(i);
	  	  	
	  	  	if (s.type==CachePoint.RP || s.type==CachePoint.PPFP)
	  	  	  if (s.timestamp==t)
	  	  	    return i;
	  	  	
	  	  	if (s.type==CachePoint.CAP || s.type==CachePoint.LRP || s.type==CachePoint.CLRP)
	  	  	  {
	  	  	  	if (i==0)
	  	  	  	  ts = C.IT;
	  	  	  	else 
	  	  	  	  ts = getStartTimestamp(C,s);	  	  	  	
	  	  	  	  
	  	  	  	te = s.timestamp;	  	  	  	
	  	  	  	
	  	  	  	if (t>=ts && t<=te)
	  	  	  	  return i;  	  	  	  	  	  	 	  	  	
	  	  	  }
	  	  	  
	  	  	if (t<getStartTimestamp(C,s))
	  	  	  return i-1;  	  	
	  	  }
	  	  
	    /*System.err.println("Cannot find the cache item with timestamp " + t +".");
	    System.err.println("C.IT = " + C.IT + "\t" + "C.attribute = " + C.attribute);
	    for (i=0;i<C.data.size();i++)
	      System.err.print(((CachePoint)C.data.elementAt(i)).timestamp + " ");
	    for (i=0;i<C.data.size();i++)
	      System.err.print(((CachePoint)C.data.elementAt(i)).type + " ");
	    System.err.println();
	    System.err.println(C.SPAN + " " + C.SP);*/
	    return 0;
	    /*System.exit(1);
	    return -1;*/
	  	  
	  }  //findCacheItem
	  
	public double getValue(CachePoint s, long t)
	  {
	  	if (s.type==CachePoint.RP)
	  	  return ((RP)s).value;
	  	else
	  	  if (s.type==CachePoint.PPFP)
	  	    return ((PPFP)s).value;
	  	  else
	  	    if (s.type==CachePoint.CAP)
	  	      return ((CAP)s).avg;
	  	    else
	  	      if (s.type==CachePoint.LRP)
	  	        return ((LRP)s).w0+((LRP)s).w1*t;
	  	      else  // if (s.type==CachePoint.CLRP)
	  	        return ((CLRP)s).w0+((CLRP)s).w1*extractCorrelatedValue(((CLRP)s).p,t);
	  }
	  
	public double extractCorrelatedValue(Cache C, long t)
	  {
	  	CachePoint s = (CachePoint)C.data.elementAt(findCacheItem(C,t));
	  	
	  	return getValue(s,t);
	  		  	
	  } //extractCorrelatedValue
	  
	public double [] constructValueArray(Cache C, EDMethod M, long ts)
	  {
	  	double S [] = new double [M.T/M.sp+1];
	  	int i = 0,j;
	  	
	  	int mark = findCacheItem(C,ts);	  	
	  	CachePoint s = (CachePoint)C.data.elementAt(mark);
	  	long t = ts;  
	  	  
	  	while (true)
	  	  {
	  	   	while (t<=s.timestamp)
	  	  	  {
	  	  	    S[i++]= getValue(s,t);
	  	  	    t+=M.sp;
	  	  	  }
	  	  	     
	  	  	if (mark==C.data.size()-1) break;
	  	  	
	  	  	s = (CachePoint)C.data.elementAt(++mark);
	  	  	//t = getStartTimestamp(C, s);	  	
	  	  }	 
	  		  	
	  	return S; 	
	  		  	
	  }  //constructValueArray
	  
	public void computeMean(RP sn, Cache C, EDMethod M, long ts)
	  {
	  	int i;
	  	
	  	if (M.type==EDMethod.HORIZON)
	  	  M = (HorizonMethod)M;
	  	else
	  	  if (M.type==EDMethod.OSCILLATION)
	  	    M = (OscillationMethod)M;
	  	  else
	  	    {
	  	      System.err.println("Error method type in Function computeMean: must be horizon or oscillation");
	  	      System.exit(1);
	  	    }  	
	  	
	  	if (M.mean==-1)
	  	  {
	  	  	double S [] = constructValueArray(C, M, ts);
	  	  	S[S.length-1] = sn.value;
	  	  	M.mean = 0;
	  	  	for (i=0;i<S.length;i++)
	  	  	  M.mean += S[i];
	  	  	M.mean /= (double)S.length;
	  	  }
	  	else
	  	  {
	  	  	double v0=getValue((CachePoint)C.data.elementAt(findCacheItem(C,sn.timestamp-M.T-M.sp)),sn.timestamp-M.T-M.sp);
	  	  		  	  	
	  	  	M.mean = (M.mean*(M.T/M.sp+1)-v0+sn.value)/(M.T/M.sp+1); 
	  	  }  	    
	  	
	  }  //computeMean
	  
	public void printCacheItems(Cache C, long ts)
	  {
	  	int mark = findCacheItem(C,ts);
	  	
	  	if (mark==-1)
	  	  {
	  	  	System.out.println("Invalid ts");
	  	  	return;
	  	  }
	  	  
	  	CachePoint s = (CachePoint)C.data.elementAt(mark);
	  	
	  	while (true)
	  	  {
	  	  	if (s.type==CachePoint.RP || s.type==CachePoint.PPFP)
	  	  	  System.out.print("(" + s.type + ", " + s.timestamp + ", " + ((RP)s).value + ")");
	  	  	else
	  	  	  {
	  	  	  	long t = getStartTimestamp(C,s);	  	  	
	  	  	    while (t<=s.timestamp)
	  	  	      {
	  	  	  	    System.out.print("(" + s.type + ", " + t + ", " + getValue(s,t) + ") ");
	  	  	  	    t+=C.SP;
	  	  	      }	
	  	  	   }  
	  	  	  
	  	  	System.out.println("");
	  	  	
	  	  	if (mark==C.data.size()-1)
	  	  	  break;	  	  	  
	  	  	else
	  	  	  s = (CachePoint)C.data.elementAt(++mark);	  	  	
	  	  }
	  	  
	  	System.out.println("C.size="+C.data.size());	
	  	    	
	  }  // printCacheItems
	
	public void generalPatternMatching(PatternMethod M, Cache C, RP sn)
	  {
	  	long te=sn.timestamp;
	  	long ts=te-M.T;
	  	
	  	M.returnTimestamp=sn.timestamp;
	  	M.returnVal=false;
	  	
	  	if (ts<C.IT) return;
	  	
	  	int count = (int)Math.ceil((M.T/M.sp+1)*(1-M.one_minus_alpha));
	  	
	  	int mark;
	  	CachePoint s = (CachePoint)C.data.elementAt(mark=findCacheItem(C,ts));	  	
	  
	  	int i=0;
	  	long t=ts;
	  	while (true)
	  	  {
	  	  	if (s.type==CachePoint.RP || s.type==CachePoint.PPFP)
	  	  	  {
	  	  	  	if (Math.abs(((RP)s).value-M.Q[i++])>(M.delta+M.epsilon))
	  	  	      count--;
	  	  	  }
	  	  	else
	  	  	  while (t<=s.timestamp)
	  	  	    {
	  	  	      if (Math.abs(getValue(s,t)-M.Q[i++])>(M.delta+M.epsilon))
	  	  	        count--;
	  	  	      if (i==M.Q.length)
	  	  	        return;
	  	  	      t+=M.sp;	
	  	  	    }   	  	
	  	  	
	  	  	if (count<0) return;
	  	  	
	  	  	if (mark==C.data.size()-1)
	  	  	  {
	  	  	  	if (Math.abs(sn.value-M.Q[i++])>(M.delta+M.epsilon)) count--;
	  	  	  	if (i==M.Q.length) return;
	  	  	  	if (count>0)
	  	  	  	  {
	  	  	  	  	M.returnVal=true;
	  	  	  	  	node.executeQuery(sn,M.q);
	  	  	  	  }
	  	  	  	
	  	  	  	break;
	  	  	  }
	  	  	
	  	  	s=(CachePoint)C.data.elementAt(++mark);
	  	  	t = getStartTimestamp(C,s);	  	  		  	  	
	  	  }
	  		  	
	  }  //generalPatternMatching
	  
	public void horizonPatternMatching(HorizonMethod M, Cache C, RP sn)
	  {
	  	long te = sn.timestamp;
	  	long ts = te - M.T;	 
	  	
	  	M.returnTimestamp=sn.timestamp;
	  	M.returnVal = false;   		  	  	
	  	
	  	if (ts<C.IT || te-M.lastTimeofMatching<=M.T)
	  	  return;
	  	  
	  	int count = (int)Math.ceil(M.n*(1-M.one_minus_alpha));	  	
	  	  	
	   	int mark;
	   	CachePoint s = (CachePoint)C.data.elementAt(mark=findCacheItem(C,ts));
	   	
	   	M.mean=-1;
		computeMean(sn,C,M,ts);	   	
	   	
	   	/*if (sn.timestamp==5190&&node.id==1)
	   	  {
	   	  	printCacheItems(C,C.IT);
	   	  	System.out.println(M.mean);
	   	  	System.exit(0);	   	  	
	   	  }*/	   	
	  	  	
	   	long t = ts;	   	
	   	while (true)
	  	  {
	  	  	if (s.type==CachePoint.RP) 
	  	  	  if (Math.abs(((RP)s).value-M.mean)>M.epsilon) count --;
	  	  	
	  	  	if (s.type==CachePoint.PPFP)
	  	  	  if (Math.abs(((PPFP)s).value-M.mean)>M.epsilon) count --;
	  	  	  
	  	  	if (s.type==CachePoint.CAP)
	  	  	  if (((CAP)s).min>M.mean || ((CAP)s).max<M.mean)
	  	  	    count-=(s.timestamp-t)/M.sp+1;
	  	  	    
	  	  	if (s.type==CachePoint.LRP)
              while (t<=s.timestamp)
                {
                  double vt=((LRP)s).w0+((LRP)s).w1*t;
                  t+=M.sp;
                  if (Math.abs(vt-M.mean)>M.epsilon/2) count--;	
                } 	  	  	   
	  	  	  
	  	  	if (s.type==CachePoint.CLRP)
	  	  	  while (t<s.timestamp)
	  	  	    {
	  	  	      double vt=((CLRP)s).w0+((CLRP)s).w1+extractCorrelatedValue(((CLRP)s).p,t);
	  	  	      t+=M.sp;
	  	  	      if (Math.abs(vt-M.mean)>M.epsilon/2) count--;		
	  	  	    }
	  	  	
	  	  	if (count<0) break;
	  	  	
	  	  	if (mark==C.data.size()-1) break;
	  	  	
	  	  	s = (CachePoint)C.data.elementAt(++mark);
	  	  	t = getStartTimestamp(C,s);	  	  		  	  	  
	  	  }
	  	  	  
	  	if (count >= 0)
	  	  {
	  	  	M.lastTimeofMatching=sn.timestamp;
	  	  	M.returnVal=true;	  	  	
	  	  	node.executeQuery(sn, M.q);
	  	  }	    
	  	
	  }  //horizonPatternMatching
	
	public void slopePatternMatching(SlopeMethod M, Cache C, RP sn)
	  {
	  	int i;
	  	long te = sn.timestamp;
	  	long ts = te-M.T;
	  	
	  	M.returnTimestamp=sn.timestamp;
	  	M.returnVal=false;	  		  	
	  	
	  	if (ts<C.IT||te-M.lastTimeofMatching<=M.T) return;
	  	
	  	CachePoint s = (CachePoint)C.data.elementAt(findCacheItem(C,ts));	  	
	  	
	  	Vector S = new Vector (10,5);
	  	if (M.S==null)
	  	  {
	  	  	double sv[] = constructValueArray(C,M,ts);
	  	  	for (i=0;i<sv.length-1;i++)
	  	  	  S.addElement(new Double(sv[i]));
	  	  } 	  
	  	else  // M.S != null
	  	  {
	  	  	S = M.S;
	  	  	
	  	  	if (!M.headEvicted) S.removeElementAt(0);
	  	  	
	  	  	double v1,v2,v3;
	  	  	v1=getValue((CachePoint)C.data.elementAt(findCacheItem(C,ts-M.sp)),ts-M.sp);
	  	  	v2=getValue((CachePoint)C.data.elementAt(findCacheItem(C,ts)),ts);
	  	  	v3=getValue((CachePoint)C.data.elementAt(findCacheItem(C,ts+M.sp)),ts+M.sp); 
	  	  	            
	  	  	if ((M.mode == EDMethod.I) && (v2 < v1) && (v2 < v3) ||
	  	  	    (M.mode == EDMethod.D) && (v2 > v1) && (v2 > v3))
	  	  	  S.insertElementAt(new Double(v2),0);  	  	   		  	  	  	
	  	  }	 	
	    
	    S.addElement(new Double(sn.value));
	    
	    int len = M.T/M.sp+1;
	    int n = S.size();
	    boolean marked;  	
	  	Vector markers = new Vector(200,200);	  	
	  
	  	while (n>=(int)Math.ceil(len*(1-2*(1-M.one_minus_alpha1))))
	  	  {
	  	  	marked = false;
	  	  	for (i=0;i<n-1;i++)	  	  	  
	  	  	  if (M.mode==EDMethod.I && ((Double)S.elementAt(i+1)).doubleValue()<((Double)S.elementAt(i)).doubleValue()
	  	  	      || M.mode==EDMethod.D && ((Double)S.elementAt(i+1)).doubleValue()>((Double)S.elementAt(i)).doubleValue())
	  	  	    {
	  	  	  	  marked = true;
	  	  	  	  if (i==0) M.headEvicted = true;
	  	  	  	  markers.addElement((Double)S.elementAt(i+1));
	  	  	  	  if (!markers.contains((Double)S.elementAt(i)))
	  	  	  	    markers.addElement((Double)S.elementAt(i));
	  	  	  	}  
	  	  	  	  	  	
	  	  	if (!marked) break;
	  	  	
	  	  	for (i=0;i<markers.size();i++)
	  	  	  S.removeElement((Double)markers.elementAt(i));	  	  		  	  	
	  	  	
	  	  	n = S.size();  	  	  	
	   	  }
	   	  
	   	if (n<(int)Math.ceil(len*(1-2*(1-M.one_minus_alpha1))))
	  	  {
	  	  	if (S.size()==0) M.S = null;
	  	  	else M.S = S;	  	  	
	  	  	return;	  
	  	  }		  	
	  	
	  	while (n>=(int)Math.ceil(len*M.one_minus_alpha2))
	  	  {
	  	  	marked = false;
	  	  	for (i=0;i<n-1;i++)
	  	  	  if (((Double)S.elementAt(i+1)).doubleValue()==((Double)S.elementAt(i)).doubleValue())
	  	  	  	{
	  	  	  	  marked = true;
	  	  	  	  if (i==0) M.headEvicted=true;	  	  	  	  	
	  	  	  	  markers.addElement((Double)S.elementAt(i));
	  	  	    } 
	  	  	     	  	  	  	  	
	  	  	if (!marked) break;
	  	  	
	  	  	for (i=0;i<markers.size();i++)
	  	  	  S.removeElement((Double)markers.elementAt(i));	  	  	
	    	n = S.size();  
	   	  } 
	   	   	
	  	if (S.size()==0)
	  	  M.S = null;
	  	else
	  	  M.S = S;		  	
	  	
	  	if (S.size()>=(int)Math.ceil(len*M.one_minus_alpha2) && Math.abs(((Double)S.elementAt(S.size()-1)).doubleValue()-((Double)S.elementAt(0)).doubleValue())>=M.delta)  // output s as a leaning change point for m
	  	  {
	  	  	M.S = null;
	  	  	M.lastTimeofMatching=sn.timestamp;
	  	  	M.returnVal=true;
	  	  	node.executeQuery(sn, M.q);
	  	  }	  	
	  	
	  }  //slopePatternMatching
	  
	public char testTrend(double mean, double value)
	  {
	  	if (value > mean)
	  	  return 'L';
	  	else
	  	  if (value == mean)
	  	    return 'E';
	  	  else  //if (value < mean)
	  	    return 'S';
	  }
	  
	public void oscillationPatternMatching(OscillationMethod M, Cache C, RP sn)
	  {
	  	long te=sn.timestamp;
	  	long ts=te-M.T;
	  	
	  	M.returnTimestamp=te;
	  	M.returnVal=false;
	  	
	  	if (ts<C.IT || te-M.lastTimeofMatching<=M.T)
	  	  return;
	  	  
	  	int count = (int)Math.ceil((M.T/M.sp+1)*M.lou*(2-M.one_minus_alpha));
	  	
	  	int mark;
	  	CachePoint s = (CachePoint)C.data.elementAt(mark=findCacheItem(C,ts));
	  	
	  	computeMean(sn,C,M,ts);
	  	
	  	char trend = 'U';
	  	char trend_org;
	  	long t=ts;
	  	while (true)
	  	  {
	  	  	trend_org = trend;	  	  
	  	  	
	  	  	if (s.type==CachePoint.RP || s.type==CachePoint.PPFP)
	  	  	  {
	  	  	  	trend=testTrend(M.mean,((RP)s).value);
	  	  	  	if (trend==trend_org) count--;
	  	  	  	else count = (int)Math.ceil((M.T/M.sp+1)*(M.lou+M.one_minus_alpha));	  	  	  	
	  	  	  }	
	  	  	  
	  	  	if (s.type==CachePoint.CAP)
	  	  	  {
	  	  	  	trend=testTrend(M.mean,((CAP)s).avg);
	  	  	  	if (trend==trend_org) count-=(s.timestamp-t)/M.sp+1;
	  	  	  	else count = (int)Math.ceil((M.T/M.sp+1)*(M.lou+M.one_minus_alpha));
	  	  	  }
	  	  	  
	  	  	if (s.type==CachePoint.LRP || s.type==CachePoint.CLRP)
	  	  	  while (t<=s.timestamp)
	  	  	    {
	  	  	  	  double vt;
	  	  	  	  if (s.type==CachePoint.LRP)
	  	  	  	    vt=((LRP)s).w0+((LRP)s).w1*t;
	  	  	  	  else
	  	  	  	    vt=((CLRP)s).w0+((CLRP)s).w1*extractCorrelatedValue(((CLRP)s).p,t);
	  	  	  	  t+=M.sp;
	  	  	  	  trend=testTrend(M.mean,vt);
	  	  	  	  if (trend==trend_org) count--;
	  	  	  	  else
	  	  	  	    {
	  	  	  	      count = (int)Math.ceil((M.T/M.sp+1)*(M.lou+M.one_minus_alpha));
	  	  	  	      trend_org = trend;
	  	  	  	    }
	  	  	    }
	  	  	    
	  	  	if (count<0) return;
	  	  	
	  	  	if (mark==C.data.size()-1)
	  	  	  {
	  	  	  	trend_org=trend;
	  	  	  	trend=testTrend(M.mean,sn.value);
	  	  	  	if (trend==trend_org) count--;
	  	  	  	else count = (int)Math.ceil((M.T/M.sp+1)*(M.lou+M.one_minus_alpha));
	  	  	  	if (count>0)
	  	  	  	  {
	  	  	  	  	M.returnVal=true;
	  	  	  	  	M.lastTimeofMatching=sn.timestamp;
	  	  	  	  	node.executeQuery(sn,M.q);
	  	  	  	  }
	  	  	  	break;
	  	  	  }
	  	  	
	  	  	s = (CachePoint)C.data.elementAt(++mark);
	  	  	t = getStartTimestamp(C,s);
	  	  		  	  	    	  	
	  	  } //while(true)	
	  }
	  
	public int checkInstantaneousChange(Cache C, EDMethod M, long ts, int n, double v)
	  {
	  	int count = (int)Math.ceil(n*(1-M.one_minus_alpha));	  
	  	
	  	int mark=findCacheItem(C,ts);
	  	CachePoint s = (CachePoint)C.data.elementAt(mark);	  		  
	  	
	  	double delta;
	  	if (M.type==EDMethod.JUMP)
	  	  delta=((JumpMethod)M).delta2;
	  	else
	  	  delta=((SpikeMethod)M).delta2;	  	
	  	
	  	while (true)
	  	  {
	  	  	if (s.type==CachePoint.RP || s.type==CachePoint.PPFP)
	  	  	  if (M.mode==EDMethod.I&&v-((RP)s).value<delta || M.mode==EDMethod.D&&v-((RP)s).value>-delta)
	  	  	    count--;
	  	  	    
	  	  	if (s.type==CachePoint.CAP)
	  	  	  if (M.mode==EDMethod.I&&v-((CAP)s).avg<delta || M.mode==EDMethod.D&&v-((CAP)s).avg>-delta)
	  	  	    count-=(s.timestamp-ts)/M.sp+1;
	  	  	    
	  	  	if (s.type==CachePoint.LRP || s.type==CachePoint.CLRP)
	  	  	  {
	  	  	  	double v1,v2;
	  	  	  	if (s.type==CachePoint.LRP)
	  	  	  	  {v1=((LRP)s).w0+((LRP)s).w1*ts; v2=((LRP)s).w0+((LRP)s).w1*s.timestamp;}
	  	  	  	else
	  	  	  	  {v1=((CLRP)s).w0+((CLRP)s).w1*extractCorrelatedValue(((CLRP)s).p,ts); v2=((CLRP)s).w0+((CLRP)s).w1*extractCorrelatedValue(((CLRP)s).p,s.timestamp);}
	  	  	  	  
	  	  	  	long t=(long)((v-delta-((LRP)s).w0)/((LRP)s).w1);  	  	  
	  	  	  	if(M.mode==EDMethod.I&&v-Math.max(v1,v2)<delta)
	  	  	  	  if (v1>v2) count-=(int)(Math.min(t,s.timestamp)-ts)/M.sp+1;
	  	  	  	  else count-=(int)(s.timestamp-Math.max(t,ts))/M.sp+1;
	  	  	  	if (M.mode==EDMethod.D&&v-Math.min(v1,v2)>-delta)
	  	  	  	  if (v1<v2) count-= (int)(Math.min(t,s.timestamp)-ts)/M.sp+1;
	  	  	  	  else count -= (int)(s.timestamp-Math.max(t,ts))/M.sp+1;	  	     	  	  	  
	  	  	  }  		  	  	  
	  	  	
	  	  	if (count<0 || mark==C.data.size()-1)
	  	  	  return count;
	  	  	
	  	  	s=(CachePoint)C.data.elementAt(++mark);
	  	  	ts=getStartTimestamp(C,s);	  	  		  	  	
	  	  }
	  	
	  } //checkInstantaneousChange 
	  
    public void checkPPFR(Cache C, EDMethod M, RP sn)
      {
      	long te=sn.timestamp-M.sp;
      	long ts;
      	int T;
      	if (M.type==EDMethod.JUMP)
      	  ts=te-(T=((JumpMethod)M).T1);
      	else
      	  ts=te-(T=((SpikeMethod)M).T1);
      	  
      	if (ts<C.IT)
      	  return;
      	  
      	double v1, v2;
      	v1 = getValue((CachePoint)C.data.elementAt(C.data.size()-1),((CachePoint)C.data.elementAt(C.data.size()-1)).timestamp);
      	v2 = sn.value;      
      	
      	double delta;
      	if (M.type==EDMethod.JUMP)
      	  delta=((JumpMethod)M).delta1;
      	else
      	  delta=((SpikeMethod)M).delta1;
      	if (M.mode==EDMethod.I && v2-v1>delta || M.mode==EDMethod.D && v2-v1<-delta)
      	  if (checkInstantaneousChange(C,M,ts,T/M.sp+1,v2)>=0)
      	    {
      	      isPPFP = true;
      	      
      	      if (ppfp==null) ppfp=new PPFP();
      	      
      	      ppfp.timestamp = sn.timestamp;
      	      ppfp.value = sn.value;
      	      
      	      ppfp.pl.addElement((EDMethod)M);
      	      
      	      /*System.out.println(node.id + " " + ppfp.timestamp + " " + ppfp.value);
      	      System.out.println(C.IT + " " + C.SP + " " + C.SPAN);
      	      System.exit(0);*/
      	    }
      	
      }  //checkPPFR  
	  
	public void jumpPatternMatching(JumpMethod M, Cache C, RP sn)
	  {
	  	checkPPFR(C,M,sn);
	  	
	  	long te=sn.timestamp;
	  	long ts=te-M.T2;
	  	
	  	M.returnTimestamp=sn.timestamp;
	  	M.returnVal=false;
	  	
	  	if (ts<C.IT) return;
	  	
	  	int mark;
	  	CachePoint s = (CachePoint)C.data.elementAt(mark=findCacheItem(C,ts));
	  	
	  	if (s.type!=CachePoint.PPFP) return;
	  	
	  	/*if (sn.timestamp==700&&node.id==1)
	  	  {printCacheItems(C,C.IT);System.exit(0);} */		  	  	 	
	  	
	  	if (M.mode==EDMethod.I) M.mode=EDMethod.D;
	  	else M.mode=EDMethod.I;	  	
	  
	  	double v1;
	  	if (mark-1 >= 0)
	  	  v1=getValue((CachePoint)C.data.elementAt(mark-1),((CachePoint)C.data.elementAt(mark-1)).timestamp);
	  	else
	  	  v1=getValue((CachePoint)C.data.elementAt(0),((CachePoint)C.data.elementAt(0)).timestamp);
	  	
	  	double v2=sn.value;
	  	
	  	int count = checkInstantaneousChange(C,M,ts,M.T2/M.sp+1,v1);  		
	  
	  	if (M.mode==EDMethod.I) M.mode=EDMethod.D;
	  	else M.mode=EDMethod.I;
	  	
	  	if (M.mode==EDMethod.I&&v2-v1<M.delta2 || M.mode==EDMethod.D&&v2-v1>-M.delta2)
	  	  count --;
	  	  
	  	if (count>0)
	  	  {
	  	  	M.returnVal=true;
	  	  	node.executeQuery(sn,M.q);
	  	  }
	  		  	
	  }  //jumpPatternMatching	  
	  
	public void spikePatternMatching(SpikeMethod M, Cache C, RP sn)
	  {
	  	checkPPFR(C,M,sn);
	  	
	  	long te=sn.timestamp;
	  	long ts=te-M.T2;
	  	
	  	M.returnTimestamp=sn.timestamp;
	  	M.returnVal=false;
	  	
	  	if (ts<C.IT) return;	  	
	  	
	  	int mark;
	  	CachePoint s = (CachePoint)C.data.elementAt(mark=findCacheItem(C,ts));
	  	
	  	if (s.type!=CachePoint.PPFP) return;	  	
	 
	  	double v1=getValue(s,s.timestamp);
	  	
	  	double v2=sn.value;
	  	
	  	int count = checkInstantaneousChange(C,M,ts,M.T2/M.sp+1,v1);	  
	  	
	  	if (M.mode==EDMethod.I&&v1-v2<M.delta2 || M.mode==EDMethod.D&&v1-v2>-M.delta2)
	  	  count --;
	  	  
	  	if (count>0)
	  	  {
	  	  	M.returnVal=true;
	  	  	node.executeQuery(sn,M.q);
	  	  	
	  	  	/*System.out.println(node.id + " " + sn.timestamp);
	  	  	System.exit(0);*/
	  	  }	  	
	  	
	  }  //spikePatternMatching
	  
	public boolean checkPLR(CachePoint s, Cache C, double w0, double w1)
	  {
	  	if (s.type!=CachePoint.RP && s.type!=CachePoint.CAP && s.type!=CachePoint.PPFP)
	  	  {
	  	  	System.err.println("Invalid type of cache point in Function checkPLR: must be an RP or a CAP.");
	  	  	System.err.println(s.type);
	  	  	System.exit(1);
	  	  }
	  	  	
	  	if (s.type==CachePoint.RP)
	  	  if (Math.abs(w0+w1*s.timestamp-((RP)s).value)<=C.E)
	  	    return true;
	  	
	  	if (s.type==CachePoint.CAP)
	  	  {
	  	  	double v1=w0+w1*getStartTimestamp(C,s);
	  	  	double v2=w0+w1*s.timestamp;
	  	  	if (Math.min(v1,v2)>=((CAP)s).min && Math.max(v1,v2)<=((CAP)s).max)
	  	  	  return true;
	  	  }
	  	  
	  	return false;
	  }  

	public void updateCache(Cache C, RP sn)
	  {
	  	if (C.data.size()==0)  // the cache is currently empty
	  	  {
	  	  	C.data.addElement((CachePoint)sn);
	  	  	C.IT = sn.timestamp;	  	  	
	  	  	return;
	  	  } 		  	
	    
	  	int mark,count;
	  	int i;
        CachePoint s;	  	
	  	if (isPPFP)
	  	  {
	  	  	if (CacheManager.PLR && C.lrs.A!=null)
	  	  	  {
	  	  	  	mark = C.data.size()-1;
	  	  	  	count = 0;
	  	  	  	Vector v = new Vector (10,5);	  	  	  	
	  	  	  	while (mark >= 0)
	  	  	  	  {
	  	  	  	 	s = (CachePoint)C.data.elementAt(mark);
	  	  	  	 	
	  	  	  	 	if (s.type==CachePoint.LRP || s.type==CachePoint.CLRP) break;
	  	  	  	  	else
	  	  	  	  	  if (s.timestamp < C.lrs.ts) break;
	  	  	  	  	  
	  	  	  	  	if (getStartTimestamp(C,s)>=C.lrs.te)
	  	  	  	  	  v.insertElementAt((CachePoint)s,0);	  	  	  	  	
	  	  	  	  	    	  	  	  	  	
	  	  	  	  	count++;
	  	  	  	  	
	  	  	  	  	mark--;	  	  	  	  	 	  	  	
	  	  	  	  }	  	  	  	
	  	  	  	
	  	  	  	for (i=0;i<count;i++)
	  	  	  	   C.data.removeElementAt(C.data.size()-1);
	  	  	  	LRP lrp = new LRP();
	  	  	  	lrp.w0 = C.lrs.w0;
	  	  	  	lrp.w1 = C.lrs.w1;
	  	  	  	lrp.timestamp = C.lrs.te;
	  	  	  	//System.out.println(lrp.timestamp + " " + ppfp.timestamp);
	  	  	  	//System.exit(0);
	  	  	    C.data.addElement((CachePoint)lrp);
	  	  	    
	  	  	    C.lrs = new LinearRegressionStructure();
	  	  	    
	  	  	    for (i=0;i<v.size();i++)
	  	  	      C.data.addElement((CachePoint)v.elementAt(i));	  	  	    
	  	  	  }
	  	  	
	  	  	C.data.addElement((CachePoint)ppfp);
	  	  	
	  	  	/*if (node.id==1&&sn.timestamp==1460)
	  	  	  {printCacheItems(C,C.IT); System.exit(0);}*/
	  	  	
	  	  	return;
	  	  }
	  	else
	  	  {
	  	  	/*if (true)
	  	      {
	  	  	    C.data.addElement((CachePoint)sn);
	  	  	    cacheEviction(C);
	  	  	    C.size = evaluateCacheSize(C);	  		  	  	
	  	  	    return;
	  	      }	*/
	  	  	
	  	  	int ungrowableMark = 0;	  	  	
	  	    
	  	    /** First-level: Piecewise Constant Approximation (PCA) */
	  	    if (CacheManager.PCA)
	  	      {
	  	        ungrowableMark = -1;
	  	        s = (CachePoint)C.data.elementAt(C.data.size()-1);
	  	        
	  	        if (s.type==CachePoint.RP)
	  	          if (Math.abs(((RP)s).value-sn.value)<=C.E)
	  	  	        {
	  	  	  	      CAP ss = new CAP();
	  	  	  	      ss.max = Math.max(sn.value,((RP)s).value);
	  	  	  	      ss.min = Math.min(sn.value,((RP)s).value);
	  	  	      	  ss.avg = (ss.max + ss.min)/2;
	  	  	  	      ss.timestamp = sn.timestamp;
	  	  	  	      C.data.removeElementAt(C.data.size()-1);
	  	  	  	      C.data.addElement((CachePoint)ss);
	  	  	        }
	  	  	      else
	  	  	        {
	  	  	  	      ungrowableMark = C.data.size() - 1;
	  	  	  	      C.data.addElement((CachePoint)sn);
	  	  	        }
	  	  	        
	  	  	    if (s.type==CachePoint.PPFP || s.type==CachePoint.LRP || s.type==CachePoint.CLRP)
	  	  	      C.data.addElement((CachePoint)sn);
	  	  	      	  	  
	  	        if (s.type==CachePoint.CAP)
	  	          {
	  	  	        if (Math.max(((CAP)s).max,sn.value)-Math.min(((CAP)s).min,sn.value)<=C.E)
	  	  	          {
	  	  	          	int n=(int)(s.timestamp-getStartTimestamp(C,s))/C.SP+1;
	  	  	          	s.timestamp=sn.timestamp;
	  	  	          	((CAP)s).avg=(((CAP)s).avg*n+sn.value)/(n+1);
	  	  	          	((CAP)s).max=Math.max(((CAP)s).max,sn.value);
	  	  	          	((CAP)s).min=Math.min(((CAP)s).min,sn.value);	  	  	          	
	  	  	          }
	  	  	        else
	  	  	         {
	  	  	         	ungrowableMark=C.data.size()-1;
	  	  	         	C.data.addElement((CachePoint)sn);
	  	  	         }
	  	          }
	  	      }  // if (CacheManager.PCA)
	  	      	  	    
	  	    
	  	    /** Second-level: Piecewise Linear Regression (PLR) */
	  	    if (!CacheManager.PLR || ungrowableMark<=0&&CacheManager.PCA)
	  	      {}
	  	    else 
	  	      {	
	  	        if (!CacheManager.PCA)
	  	          {
	  	          	C.data.addElement((CachePoint)sn); // insert the new RP into the cache
	  	          	
	  	          	ungrowableMark = C.data.size()-1;
	  	          }
	  	          
	  	        if (C.data.size()==1 || ungrowableMark<=0) return;
	  	        
	  	        s = (CachePoint)C.data.elementAt(ungrowableMark);
	  	        CachePoint ss = (CachePoint)C.data.elementAt(ungrowableMark-1);
	          	
	          	if (ss.type==CachePoint.LRP || ss.type==CachePoint.CLRP || ss.type==CachePoint.PPFP)
	          	  {}  // no operation	          	
	          	else  // ss is an RP or a CAP
	          	  {
	          	  	double w0,w1;
	          	  	
	          	  	if (C.lrs.A == null)  // C.R == null
	  	              {
	  	  	            /** 
	  	  	             * Linear Regression:
	  	         	     * h1(t) = 1, h2(t) = t
	  	          	     * H = (h1(t1), h2(t1)     H' = (h1(t1), h1(t2)    f = (f(t1) = v1
	  	          	     *      h1(t2), h2(t2))         h2(t1), h2(t2))        f(t2) = v2)
	  	          	     * A = H'H = (h1(t1)*h1(t1)+h1(t2)*h1(t2), h1(t1)*h2(t1)+h1(t2)*h2(t2)  = (2, t1+t2
	  	         	     *            h2(t1)*h1(t1)+h2(t2)*h1(t2), h2(t1)*h2(t1)+h2(t2)*h2(t2))    t1+t2, t1*t1+t2*t2)
	  	          	     * b = H'f = (h1(t1)*v1+h1(t2)*v2  = (v1+v2
	  	         	     *            h2(t1)*v1+h2(t2)*v2)    t1*v1+t2*v2)
	  	           		 * w = (w0
	  	  	        	 *      w1)
	  	  	         	* Aw = b
	  	  	         	**/	  	  	         
	  	  	         	long t1,t2;	  	  	         	
	  	  	         	double v1, v2;
	  	  	         	t1=(getStartTimestamp(C,ss)+ss.timestamp)/2;
	  	  	         	t2=(getStartTimestamp(C,s)+s.timestamp)/2;
	  	  	         	v1=getValue(ss,ss.timestamp);
	  	  	         	v2=getValue(s,s.timestamp);
	  	  	            
	  	  	         	double A [][] = {{2, t1+t2}, {t1+t2, t1*t1+t2*t2}};
	  	  	         	double b [] = {v1+v2, t1*v1+t2*v2};	  	  	         	
	  	  	         	w1 = (b[1] - b[0]*A[1][0]/A[0][0])/(A[1][1]-A[0][1]*A[1][0]/A[0][0]);
	  	  	         	w0 = (b[0] - A[0][1]*w1)/A[0][0]; 	  	         	
	  	  	      
	  	  	         	if (checkPLR(ss,C,w0,w1)&&checkPLR(s,C,w0,w1))
	  	  	         	  {
	  	  	         	  	C.lrs.A = A;
	  	  	         		C.lrs.b = b;
	  	  	         		C.lrs.w0 = w0;
	  	  	         		C.lrs.w1 = w1;
	  	  	         		C.lrs.ts = getStartTimestamp(C,ss);	  	  	         		
	  	  	         		C.lrs.te = s.timestamp;
	  	  	         		C.lrs.W = C.lrs.te - C.lrs.ts;
	  	  	         		
	  	  	         		/*System.out.println(C.lrs.ts + " " + " " + C.lrs.te + " " + C.IT);
	  	  	         		System.out.println(ss.type + " " + s.type);
	  	  	         		System.exit(0);*/	  	         		
	  	  	         	  }
	  	  	         	else return;  	  	   	  	  	
	  	          	  }
	  	        	else  // C.R != null
	  	          	  {
	  	  	        	double A [][] = {{C.lrs.A[0][0],C.lrs.A[0][1]}, {C.lrs.A[1][0], C.lrs.A[1][1]}};
	  	  	        	double b [] = {C.lrs.b[0], C.lrs.b[1]};
	  	  	 
	  	  	            long t;
	  	  				double v;
	  	  				if (s.type==CachePoint.RP)
	  	  	  		      {
	  	  	  				t = s.timestamp;
	  	  	  				v = ((RP)s).value;
	  	      		      }
	  	  				else
	  	  	              {
	  	  	  				t = (getStartTimestamp(C,s) + s.timestamp)/2;
	  	  	  				v = ((CAP)s).avg;
	  	                  }	
	  	  	
	  	  				A[0][0] += 1;  A[0][1] += t;
	  	  				A[1][0] += t;  A[1][1] += t*t;	  	  	
	  	  				b[0] += v; b[1] += t*v;	  	  				
	  	  				w1 = (b[1] - b[0]*A[1][0]/A[0][0])/(A[1][1]-A[0][1]*A[1][0]/A[0][0]);
	  	  				w0 = (b[0] - A[0][1]*w1)/A[0][0];
	  	  	
	  	  				boolean cont=true;
	  	  				
	  	  				mark = ungrowableMark;
	  	  				ss = (CachePoint)C.data.elementAt(mark);	  	  				
	  	  				while (true)
	  	  	  		  	  {
	  	  	  				if (getStartTimestamp(C,ss)<C.lrs.ts) break;
	  	  	  				
	  	  	  				cont = checkPLR(ss,C,w0,w1);
	  	  	  				
	  	  	  				/*if (mark==0)
	  	  	  				  {
	  	  	  				  	System.out.println(C.IT + " " + getStartTimestamp(C,ss) + " " + C.lrs.ts);
	  	  	  				  	System.exit(0);
	  	  	  				  }*/	  	  				
	  	  	  				
	  	  	  				if (!cont || getStartTimestamp(C,ss)==C.lrs.ts&&mark==0) break;
	  	  	  				else ss=(CachePoint)C.data.elementAt(--mark);	  	  	  				
	  	  	 		      }
	  	  	 		      
	  	  	 		    if (cont)
	  	  	 		      {
	  	  	 		      	C.lrs.te=s.timestamp;
	  	  	 		      	C.lrs.A = A;
	  	  	 		      	C.lrs.b = b;
	  	  	 		      	C.lrs.w0=w0;
	  	  	 		      	C.lrs.w1=w1;
	  	  	 		      	C.lrs.W=C.lrs.te-C.lrs.ts;
	  	  	 		      }	  	
	  	  		
	  	  				if (!cont || C.lrs.te-C.lrs.ts>LinearRegressionStructure.Wmax)
	  	  	  		      {
	  	  	  				for (i=C.data.size()-1;i>=0;i--)
	  	  	  				  {
	  	  	  				  	ss=(CachePoint)C.data.elementAt(i);
	  	  	  				  	if (getStartTimestamp(C,ss)<C.lrs.ts) break;
	  	  	  				  	else 
	  	  	  				  	  if (ss.timestamp<=C.lrs.te)
	  	  	  				  	    C.data.removeElementAt(i);	  	  	  				  	
	  	  	  				  }	  	  	  				  
	  	  	  				  	  	  	  				
	  	  	  				LRP lrp = new LRP();
	  	  	  				lrp.w0 = C.lrs.w0;
	  	  	  				lrp.w1 = C.lrs.w1;
	  	  	  				lrp.timestamp = C.lrs.te;
	  	  	  				
	  	  	    			boolean inserted = false;
	  	  	    			for (i=C.data.size()-1;i>=0;i--)
	  	  	    			  {
	  	  	    			  	ss=(CachePoint)C.data.elementAt(i);
	  	  	    			  	if (lrp.timestamp>ss.timestamp)
	  	  	    			  	  {C.data.insertElementAt((CachePoint)lrp,i+1); inserted = true; break;}
	  	  	    			  }
	  	  	    			if (!inserted) C.data.insertElementAt((CachePoint)lrp,0); 	  	    			
	  	  	    			
	  	  	  				C.lrs = new LinearRegressionStructure(); 	  	  	
	  	  	  		      }	  	  			  	  	
	  	              } 
	  	            }	  	         	  	      
	  	      }
	  	  }	  	 
	  	
	  	cacheEviction(C);
	  	
	  	C.size = evaluateCacheSize(C); 		
	  	
	  	int totalCacheSize = 0;
	  	for (i=0;i<node.caches.size();i++)
	  	  totalCacheSize += ((Cache)node.caches.elementAt(i)).size;
	  	if (totalCacheSize > node.CACHE_LIMIT)
	  	  exceedCacheSpace(C);
	  	
	  	for (i=0;i<node.caches.size();i++)
	  	   ((Cache)node.caches.elementAt(i)).size = evaluateCacheSize((Cache)node.caches.elementAt(i));
	  	  
	  	C.size = evaluateCacheSize(C);
	  	totalCacheSize = C.size;
	  	while (totalCacheSize > node.CACHE_LIMIT)
	  	  {
	  	  	C.size -= ((CachePoint)C.data.elementAt(0)).size;
	  	  	totalCacheSize = C.size;
	  	  	C.data.removeElementAt(0);
	  	  } 	  
	  	
	  }  //updateCache
	  
	public int evaluateCacheSize(Cache C)
	  {
	  	int i,size = 0;
	  	for (i=0;i<C.data.size();i++)
	  	  {
	  	  	CachePoint s = (CachePoint)C.data.elementAt(i);
	  	  	
	  	  	size+=s.size;
	  	  }
	  	  
	    return size;
	  	
	  } //evaluateCacheSize	  
	  
	public void cacheEviction(Cache C)
	  {
	  	long t = ((CachePoint)C.data.elementAt(C.data.size()-1)).timestamp;
	  	
	  	if (t-C.SPAN>=0)
	  	  C.IT = t - C.SPAN;
	  	else
	  	  C.IT = 0;
	  	
	  	CachePoint s = (CachePoint)C.data.elementAt(0);	  	
	  	if (s.timestamp < C.IT)
	  	  C.data.removeElementAt(0);
	  	  
	  	if (C.lrs.A != null)
	  	  {
	  	  	if (C.lrs.ts < C.IT)
	  	  	  C.lrs.ts = C.IT;
	  	  	  
	  	  	if (C.lrs.te < C.IT)
	  	  	  C.lrs = new LinearRegressionStructure();	  	  	
	  	  }  
	  	  	  	
	  }  //cacheEviction 
	  
	  
    public void exceedCacheSpace (Cache C)
      {
      	int i;        	   	
      	byte tag;
      	
      	/*if (node.id==1)
      	  System.out.println(evaluateCacheSize(C));*/
      	
      	if (examineCacheStatus(C)==1) return;  // examine C first

      	for (i=0;i<node.caches.size();i++) // examine other caches
      	  {
      	  	if (((Cache)node.caches.elementAt(i))==C) continue;
      	  	
      	  	tag = examineCacheStatus((Cache)node.caches.elementAt(i));
      	  	if (tag == 1) return;
      	  } 	  
      	
      	Cache sel_C = (Cache)node.caches.elementAt(0);
      	double minCost = sel_C.minCost;
      	for (i=1;i<node.caches.size();i++)
      	  {
      	  	if (minCost > ((Cache)node.caches.elementAt(i)).minCost)
      	  	  {
      	  	  	sel_C = (Cache)node.caches.elementAt(i);
      	  	  	minCost = sel_C.minCost;
      	      }
      	      
      	    if (minCost == ((Cache)node.caches.elementAt(i)).minCost)
      	      if (((Cache)node.caches.elementAt(i)).equals(C))
      	        sel_C = (Cache)node.caches.elementAt(i);
      	  }
      	
        if (minCost != Double.MAX_VALUE)
          compressCache(sel_C);          
        
        /*if (node.id==1)
          {System.out.println(evaluateCacheSize(C)); System.exit(0);}*/
      	      	
      }  //exceedCacheSpace
      
    public int constructTimeSeries(Cache C, CachePoint s, int i, long ts, long te, long t [], double v [])
      {
      	if (s.type==CachePoint.RP)
	  	  {
	  	  	 if (i>=t.length)
	  	  	   return i;
	  	  	 t[i] = s.timestamp;
	  	  	 v[i++] = ((RP)s).value;
	  	   }
	  	else
	  	  if (s.type==CachePoint.CAP)
	  	    for (int j=0;j<(te-ts)/C.SP+1;j++)
	  	  	  {
	  	  	    if (i>=t.length)
	  	  	      return i;
	  	  	    
	  	  	    t[i] = ts + j*C.SP;
	  	  	    v[i++] = ((CAP)s).avg;	  	  	    
	  	  	   }
	  	  else
	  	    if (s.type==CachePoint.LRP || s.type==CachePoint.CLRP)
	  	      {
	  	  	    long tt = ts;
	  	  	    while (tt <= te)
	  	  	      {
	  	  	       	if (i>=t.length)
	  	  	       	  return i;
	  	  	       	
	  	  	       	t[i]= tt;
	  	  	       	if (s.type==CachePoint.LRP)
	  	  	       	  v[i++] = ((LRP)s).w0 + ((LRP)s).w1*tt;
	  	  	       	else
	  	  	       	  v[i++] = ((CLRP)s).w0 + ((CLRP)s).w1*extractCorrelatedValue(((CLRP)s).p,tt);
	  	  	       	tt += C.SP;	  	  	       	
	  	  	      }
	  	  	   }
	  	  	   
	  	return i;
      	
      } //constructTimeSeries
	  
	public byte examineCacheStatus (Cache C)
	  {
	  	if (C.data.size() == 0)
	  	  {
	  	  	C.minCost = Double.MAX_VALUE;
	  	  	return 2;  // indicates that the cache cannot be compressed
	  	  }
	  	
	  	
	  	CachePoint s;
	  	CachePoint s1, s2, s3, s4;
	  	int i,j;
	  	long ts1, ts2, te1, te2;
	  	if (C.lrs.A != null)
	  	  {
	  	  	while (true)
	  	  	  {
	  	  	  	if (C.data.size() == 0) break;
	  	  	  		  	  	  	
	  	  	  	s = (CachePoint)C.data.elementAt(C.data.size()-1);	  	  	  
	  	  	  	
	  	  	  	if (s.timestamp >= C.lrs.ts)
	  	  	  	  C.data.removeElementAt(C.data.size()-1);
	  	  	  	else
	  	  	  	  break;
	  	  	  }
	  	  	
	  	  	LRP lrp = new LRP();
	  	  	lrp.w0 = C.lrs.w0;
	  	  	lrp.w1 = C.lrs.w1;
	  	  	lrp.timestamp = C.lrs.te;
	  	  	C.data.addElement((CachePoint)lrp);
	  	  	
	  	  	C.lrs = new LinearRegressionStructure();
	  	  	
	  	  	return 1;	// indicate that no need for "lossy" LRP merging  	  	
	  	  }	  	
	  	  
	  	if (CacheManager.PLR)
	  	  {
	  	  	/** Check the first two cache items for merge */	  	  	
	  	  	int markf = 0;  	  	
	  	  	while (true)
	  	  	  {
	  	  	  	if (markf > C.data.size() - 2)
	  	  	      {
	  	  	      	C.minCost = Double.MAX_VALUE;
	  	  	  		return 2;
	  	  	  	  }
	  	  	      	  
	  	  	  	s1 = (CachePoint)C.data.elementAt(markf);	  	  	  	
	  	  	  	if (s1.type==CachePoint.PPFP) {markf++; continue;}	  	  	  	
	  	  	    s2 = (CachePoint)C.data.elementAt(markf+1);	  	  	    
	  	  	    if (s2.type==CachePoint.PPFP) {markf+=2; continue;}
	  	  	    
	  	  	    break;         
	  	  	    
	  	      }  // while (true)  	  	
	  	  	
	  	  	if (markf == 0)
	  	  	  ts1 = C.IT;
	  	  	else
	  	  	  ts1 = getStartTimestamp(C,s1);
	  	  	te1 = s1.timestamp;
	  	  	ts2 = getStartTimestamp(C,s2);
	  	  	te2 = s2.timestamp;
	  	  	
	  	  	int num = (int)((te1-ts1)/C.SP + 1 + (te2-ts2)/C.SP + 1);
	  	  	if (num <= 0)
	  	  	  return 2;
     	  	double v [] = new double [num];
	  	  	long t [] = new long [num];
	  	  	i = 0;
	  	  	i=constructTimeSeries(C,s1,i,ts1,te1,t,v);
	  	  	i=constructTimeSeries(C,s2,i,ts2,te2,t,v);	  	  	
	  	  	
	  	  	double regressionData [][] = new double [num][2];
	  	  	for (i=0;i<num;i++)
	  	  	  {
	  	  	  	regressionData[i][0] = v[i];
	  	  	  	regressionData[i][1] = t[i];
	  	  	  }
	  	  	
	  	  	double cf [] = LinearRegression.linearEquation(regressionData, 1);
	  	  	double cost1 = Math.abs(cf[0] + cf[1]*t[0] - v[0]);
	  	  	for (i=1;i<num;i++)
	  	  	  if (Math.abs(cf[0] + cf[1]*t[i] - v[i]) > cost1)
	  	  	    cost1 = Math.abs(cf[0] + cf[1]*t[i] - v[i]);	  	  	
	  	  
	  	  	/** Check the last two cache items for merge */
	  	  	int markl = C.data.size() - 1;	  	  		  	    	  	
	  	  	while (true)
	  	  	  {
	  	  	  	if (markl < 1)
	  	  	      {
	  	  	      	C.minCost = Double.MAX_VALUE;
	  	  	      	return 2;
	  	  	      }
	  	  	      
	  	  	    s4 = (CachePoint)C.data.elementAt(markl);
	  	  	    if (s4 == null)
	  	  	      return 2;	  	  	  	
	  	  	  	if (s4.type==CachePoint.PPFP) {markl--; continue;}	  	  	  	
	  	  	    s3 = (CachePoint)C.data.elementAt(markl-1);	  	  	    
	  	  	    if (s3.type==CachePoint.PPFP) {markl-=2; continue;}
	  	  	    
	  	  	    break; 	  	
	  	      }	  	  	
	  	    
	  	  	if (markl < 2)
	  	  	  ts1 = C.IT;
	  	  	else
	  	  	  ts1 = getStartTimestamp(C,s3);
	  	  	te1 = s3.timestamp;
	  	  	ts2 = getStartTimestamp(C,s4);
	  	  	te2 = s4.timestamp;
	  	  	
	  	  	num = (int)Math.ceil((te1-ts1)/C.SP + 1 + (te2-ts2)/C.SP + 1);
	  	  	if (num <= 0)
	  	  	  return 2;
	  	  	v = new double [num];
	  	  	t = new long [num];
	  	  	i = 0;
	  	  	i=constructTimeSeries(C,s3,i,ts1,te1,t,v);
	  	  	i=constructTimeSeries(C,s4,i,ts2,te2,t,v);  	  
	  	  	
	  	  	regressionData = new double [num][2];
	  	  	for (i=0;i<num;i++)
	  	  	  {
	  	  	  	regressionData[i][0] = v[i];
	  	  	  	regressionData[i][1] = t[i];
	  	  	  }
	  	  	
	  	  	double cl [] = LinearRegression.linearEquation(regressionData, 1);
	  	  	double cost2 = Math.abs(cl[0] + cl[1]*t[0] - v[0]);
	  	  	for (i=1;i<num;i++)
	  	  	  if (Math.abs(cl[0] + cl[1]*t[i] - v[i]) > cost2)
	  	  	    cost2 = Math.abs(cl[0] + cl[1]*t[i] - v[i]); 	  	 	
	  	  	
	  	  	double minCost;
	  	  	if (cost1 <= cost2)
              minCost = cost1;
            else
              minCost = cost2;
                       
	  	  	if (C.size == 0)
	  	  	  minCost = Double.MAX_VALUE;
	  	  	else
	  	  	  minCost /= C.size;  	  	
	  	  	  
	  	  	C.cost1 = cost1; C.cost2 = cost2;
	  	  	C.s1 = s1; C.s2 = s2; C.s3 = s3; C.s4 = s4;
	  	  	C.markf = markf; C.markl = markl;
	  	  	C.cf = cf; C.cl = cl;
	  	  	C.minCost = minCost; 	  	
	  	  
	  	  	return 0;	 
	  	  	 	  		  	  	
	  	  }  // if (CacheManager.PLR)
	  	  
	  	if (CacheManager.PCA)
	  	  {
	  	  	/** Check the first two cache items for merge */	  	  	
	  	  	int markf = 0;	  	  	
	  	  	while (true)
	  	  	  {
	  	  	  	if (markf > C.data.size() - 2)
	  	  	      {
	  	  	      	C.minCost = Double.MAX_VALUE;
	  	  	  		return 2;
	  	  	  	  }
	  	  	      	  
	  	  	  	s1 = (CachePoint)C.data.elementAt(markf);	  	  	  	
	  	  	  	if (s1.type==CachePoint.PPFP) {markf++; continue;}	  	  	  	
	  	  	    s2 = (CachePoint)C.data.elementAt(markf+1);	  	  	    
	  	  	    if (s2.type==CachePoint.PPFP) {markf+=2; continue;}
	  	  	    
	  	  	    break;  	  	    
	  	      } 
	  	  	
	  	  	if (markf == 0)
	  	  	  ts1 = C.IT;
	  	  	else
	  	  	  ts1 = getStartTimestamp(C,s1);
	  	  	te1 = s1.timestamp;
	  	  	ts2 = getStartTimestamp(C,s2);
	  	  	te2 = s2.timestamp;
	  	  	
	  	  	int num = (int)((te1-ts1)/C.SP + 1 + (te2-ts2)/C.SP + 1);
	  	  	double v [] = new double [num];
	  	  	long t [] = new long [num];
	  	  	i = 0;
	  	  	i=constructTimeSeries(C,s1,i,ts1,te1,t,v);
	  	  	i=constructTimeSeries(C,s2,i,ts2,te2,t,v);
	  	  	
	  	  	double maxf = v[0], minf = v[0], avgf=v[0];
	  	  	for (i=1;i<num;i++)
	  	  	  {
	  	  	  	if (v[i] < minf) minf = v[i];
	  	  	  	if (v[i] > maxf) maxf = v[i];
	  	  	  	avgf += v[i];
	  	  	  }
	  	  	avgf /= num;
	  	  	double cost1 = Math.abs(maxf - minf - C.E);
	  	  	
	  	  	/** Check the last two cache items for merge */
	  	  	int markl = C.data.size() - 1;	  	  		
	  	  	while (true)
	  	  	  {
	  	  	  	if (markl < 1)
	  	  	      {
	  	  	      	C.minCost = Double.MAX_VALUE;
	  	  	      	return 2;
	  	  	      }
	  	  	      
	  	  	    s4 = (CachePoint)C.data.elementAt(markl);	  	  	  	
	  	  	  	if (s4.type==CachePoint.PPFP) {markl--; continue;}	  	  	  	
	  	  	    s3 = (CachePoint)C.data.elementAt(markl-1);	  	  	    
	  	  	    if (s3.type==CachePoint.PPFP) {markf-=2; continue;}
	  	  	    
	  	  	    break; 	
	  	      }	  	  	
	  	  	
	  	  	if (markl < 2)
	  	  	  ts1 = C.IT;
	  	  	else
	  	  	  ts1 = getStartTimestamp(C,s3);
	  	  	te1 = s3.timestamp;
	  	  	ts2 = getStartTimestamp(C,s4);
	  	  	te2 = s4.timestamp;
	  	  	
	  	  	num = (int)Math.ceil((te1-ts1)/C.SP + 1 + (te2-ts2)/C.SP + 1);
	  	  	v = new double [num];
	  	  	t = new long [num];
	  	  	i = 0;
	  	  	i=constructTimeSeries(C,s3,i,ts1,te1,t,v);
	  	  	i=constructTimeSeries(C,s4,i,ts2,te2,t,v);  	  
	  	  		  	  	         
	  	  	double maxl = v[0], minl = v[0], avgl=v[0];
	  	  	for (i=1;i<num;i++)
	  	  	  {
	  	  	  	if (v[i] < minl) minl = v[i];
	  	  	  	if (v[i] > maxl) maxl = v[i];
	  	  	  	avgl += v[i];
	  	  	  }
	  	  	avgl /= num;
	  	  	double cost2 = Math.abs(maxl - minl - C.E); 
	  	  	
	  	  	double minCost;
	  	  	if (cost1 <= cost2)
              minCost = cost1;
            else
              minCost = cost2;
                       
	  	  	if (C.size == 0)
	  	  	  minCost = Double.MAX_VALUE;
	  	  	else
	  	  	  minCost /= C.size;
	  	  	  
	  	  	C.cost1 = cost1; C.cost2 = cost2;
	  	  	C.s1 = s1; C.s2 = s2; C.s3 = s3; C.s4 = s4;
	  	  	C.markf = markf; C.markl = markl;
	  	  	C.maxf = maxf; C.avgf = avgf; C.minf = minf;
	  	  	C.maxl = maxl; C.avgl = avgl; C.minl = minl;
	  	  	C.minCost = minCost;	  	
	  	  	
	  	  	return 0;
	  	  		  	  		  	  	
	  	  }  // if (CacheManager.PCA) 
	  	  
	  	return 2; 	
	  	
	  }  //examineCacheStatus
	  
	public void compressCache (Cache C)
	  {
	  	 if (CacheManager.PLR)
	  	   {
	  	   	  LRP lrp = new LRP();
	  	  	  if (C.cost1 <= C.cost2)
	  	  	    {
	  	  	  	  lrp.w0 = C.cf[0];
	  	  	  	  lrp.w1 = C.cf[1];
	  	  	  	  lrp.timestamp = C.s2.timestamp;
	  	  	  	
	  	  	  	  C.data.removeElement(C.s1);
	  	  	  	  C.data.removeElement(C.s2);
	  	  	  	  C.data.insertElementAt((CachePoint)lrp, C.markf);  	  	  	  
	  	  	    }
	  	  	  else
	  	  	    {
	  	  	  	  lrp.w0 = C.cl[0];
	  	  	  	  lrp.w1 = C.cl[1];
	  	  	  	  lrp.timestamp = C.s4.timestamp;
	  	  	  	
	  	  	  	  C.data.removeElement(C.s3);
	  	  	  	  C.data.removeElement(C.s4);
	  	  	  	  if (C.markl-1>C.data.size()-1)
	  	  	  	    C.data.insertElementAt((CachePoint)lrp,0);
	  	  	  	  else
	  	  	  	    C.data.insertElementAt((CachePoint)lrp, C.markl-1);  	  	  	  
	  	  	    }	  	  	  	 
	  	  	  
	  	  	  return;	  	   	
	  	   }
	  	   
	  	 if (CacheManager.PCA)
	  	   {
	  	   	  CAP cap = new CAP();
	  	  	  if (C.cost1 <= C.cost2)
	  	  	    {
	  	  	  	  cap.min = C.minf;
	  	  	  	  cap.max = C.maxf;
	  	  	  	  cap.avg = C.avgf;
	  	  	  	  cap.timestamp = C.s2.timestamp;
	  	  	  	
	  	  	      C.data.removeElement(C.s1);
	  	  	  	  C.data.removeElement(C.s2);
	  	  	  	  C.data.insertElementAt((CachePoint)cap, C.markf);	  	  	  	  
	  	  	    }
	  	  	  else
	  	  	    {
	  	  	  	  cap.min = C.minl;
	  	  	  	  cap.max = C.maxl;
	  	  	  	  cap.avg = C.avgl;
	  	  	  	  cap.timestamp = C.s4.timestamp;
	  	  	  	
	  	  	  	  C.data.removeElement(C.s3);
	  	  	  	  C.data.removeElement(C.s4);
	  	  	  	  C.data.insertElementAt((CachePoint)cap, C.markl-1);  	  	  	  
	  	  	    }  	  	    
	  	  	  
	  	   	  return;
	  	   }
	  	
	  }  //compressCache	
} 