/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/** 
 * A query plan structure installed on a node.
 *
 * @see Node
 * @see Educt
 */
public class Query {
	
	int id;  // id of the query
	
	String queryText;  // SQL text of the query
	
	String attributes [] = new String [Educt.attributes.length];
	
	int attrNum = 0;
	
	EDMethod M;  // event detection method embedded in the query
	
	static final byte EVENT_DRIVEN = 0x01;
	static final byte THRESHOLD = 0x02;
	byte type;
	
	Vector predicates = new Vector (10,5);
		
	static final String logPathPrefix = "../logs/";	
	String logDirName;
	
	static int DEFAULT_SAMPLE_PERIOD = 1;  // unit: second
	
	/** Working variables for performance logging */
	double maxCT,minCT,sumCT,singleCT,tsCT,teCT;
	double maxSR,minSR,sumSR;
	int count;
	double traffic,eventNo;	
	
	Query (int id, String queryText) 
	  {
	  	this.id = id;
	  	
	  	this.queryText = queryText;	  	
	  
	   	// generate the directory to store the logs of nodes for this query
	   	logDirName = logPathPrefix + "Q"+ String.valueOf(id) + "/";	  	
	  	try 
	  	  {
	  	  	File logDir = new File(logDirName); 
	  	  	logDir.mkdirs();
	  	  }
	  	catch (Exception e) {e.printStackTrace();}  	
	  }	
	  
   public Query copy()
     {
       Query q = new Query(this.id,this.queryText);
       int i;
       
       q.attrNum = this.attrNum;
       for (i=0;i<this.attrNum;i++)
         q.attributes[i]=this.attributes[i];
       if (M != null)
         {
           if (M.type==EDMethod.PATTERN)
             q.M=((PatternMethod)M).copy();
           else
             if (M.type==EDMethod.HORIZON)
               q.M=((HorizonMethod)M).copy();
             else
               if (M.type==EDMethod.SLOPE)
                 q.M=((SlopeMethod)M).copy();
               else
                 if (M.type==EDMethod.OSCILLATION)
                   q.M=((OscillationMethod)M).copy();
                 else
                   if (M.type==EDMethod.JUMP)
                     q.M=((JumpMethod)M).copy();
                   else
                     if (M.type==EDMethod.SPIKE)
                       q.M=((SpikeMethod)M).copy();
           q.M.q = q;
         }
       q.type=this.type;       
       for (i=0;i<this.predicates.size();i++)
         q.predicates.addElement(((Predicate)this.predicates.elementAt(i)).copy());
       q.logDirName=this.logDirName;
       
       q.maxCT=this.maxCT; q.minCT=this.minCT; q.sumCT=this.sumCT;
       q.maxSR=this.maxSR; q.minSR=this.minSR; q.sumSR=this.sumSR;
       q.count=this.count;
       q.traffic=this.traffic; q.eventNo=this.eventNo;
       
       return q;    	
     }	
}