/*

* Copyright (c) 2006, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;

public class Sensor {
	
	Node node;	
	String attribute;
	
	String dataFile; // indicates the file where this sensor should read data from
	String dataFilePathPrefix = "../conf/system/data/";
	BufferedReader dataReader;
	
	static final int DATASET_SAMPLE_PERIOD=31; // the sample period of the sensory data set used (Intel Lab, Pervasive Lab or Frog Pond)
	
	Sensor (Node node, String attribute)
	  {
	  	this.attribute = attribute;
	  			  	
	  	this.node = node;  	
	  	
	  	dataFile = dataFilePathPrefix + attribute + "/node" + node.id + ".dat";	  	
	  	
	  	try {dataReader = new BufferedReader(new FileReader(dataFile));}
	  	catch (Exception e) {e.printStackTrace();}	  	
	  }
	  
    public String sample(long startTimestamp, int samplePeriod)
      {
      	
      	/** 
      	 * We assume the data file has a format of: time  value.
      	 * time is an integer with unit second. 
      	 */
      	try
      	  {
      	  	while (true)
      	  	  {
      	  	  	String line = dataReader.readLine();
      	  	
      	  	    if (line == null)
      	  	      {
      	  	  	    dataReader.close();      	  	  	    
      	  	  	    return null;      	  	  	
      	          }      	    
      	      	
      	      	double value = Double.MAX_VALUE;
      	      	long endTimestamp = 0;
      	      	int mark = line.indexOf(" ");
      	      	if (mark == -1)
      	      	  mark = line.indexOf("\t");
      	      	if (mark == -1)
      	      	  {
      	  	  	  	System.err.println("Data file " + dataFile + " format error! Please check.");
      	  	  	  	System.exit(0);
      	  	  	  }
      	  	  	String str = line.substring(0,mark).trim();
      	  	  	try {endTimestamp = Long.parseLong(str);}
      	  	  	catch (Exception e) {System.err.println("Data file " + dataFile + " format error! Please check."); System.exit(0);}                     
      	      	if (endTimestamp - startTimestamp < samplePeriod)
      	      	  continue;
      	      	str = line.substring(mark);      	  	  	 
      	      	try {value = Double.parseDouble(str);}
      	      	catch (Exception e) {System.err.println("Data file " + dataFile + " format error! Please check."); System.exit(0);}                     
      	      	return line;
      	      }	  	
      	  }
      	catch (Exception e) {e.printStackTrace();} 
      	
      	return null;     	
      	
      }  // sample
}